# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.


usage_criteria = {
  "Any Activity": "is_active",
  "5 URI": "is_active_active",
  "New Profile": "new_profile",
  "New Profile Requiring Ping on Same Day": "new_profile_old",
}
